#include <atmel_start.h>
#include <stdlib.h>

#include <COAST.h>
#include <i2c_slave.h>
#include <flash_utils.h>
#include <float_utils.h>
#include <usb_debug.h>

const char MFR_ID[] = "CERN (BE/CO)";
const char MFR_MDL[] = "DI/OT MoniMod";
const char MFR_REV[] = "0.1p";
const char MFR_LOC[] = "Geneva";
const char MFR_DAT[] = "190801";
const char MFR_SER[] = "123456789";

void page_chk();
uint16_t page __xMR;

extern uint16_t use_pec __xMR;
uint8_t use_pec_local;
void set_pec();

void accvolt();
void acccurr();
void accpowr();
uint16_t curpage_volt;
uint16_t curpage_curr;
uint16_t curpage_powr;

void query_prp();
void query_cln();
uint8_t query_r;

extern uint16_t temps_lin[3];
extern uint16_t volts_lin[3];
extern uint16_t currs_lin[3];
extern uint16_t powrs_lin[3];
extern uint16_t frpms_lin[3];

uint32_t TMR_ERROR_CNT;

static uint8_t dummy_byte;

void set_frpms();
uint16_t setfrpms_lin[3];
extern uint16_t setfrpms[3] __xMR;

void set_tc_curve();
uint16_t temp_curve_points_data[7];
extern float temp_curve_points_x[3][3] __xMR;
extern float temp_curve_points_y[3][3] __xMR;

void set_tc_matrix();
void set_tc_onoff();
uint16_t temp_matrix_row[3]; // these rows mix the 3 temps separately for each fan
uint8_t tc_on;
extern float temp_matrix[3][3] __xMR;
extern uint16_t temp_control_on __xMR;

uint8_t fan_config_1_2;
uint8_t fan_config_3_4;
void fan_config();
extern uint16_t fan_installed[3] __xMR;
extern uint16_t fan_cmdrpm[3] __xMR;
extern uint16_t fan_ppr[3] __xMR;

static const int8_t cmd_data_lengths[] = {
        1,                         // 0x00
        -1,                        // 0x1A
        2,                         // 0x8B
        2,                         // 0x8C
        2,                         // 0x8D
        2,                         // 0x8E
        2,                         // 0x8F
        2,                         // 0x90
        2,                         // 0x91
        2,                         // 0x92
        2,                         // 0x96
        -(int8_t)sizeof(MFR_ID),   // 0x99
        -(int8_t)sizeof(MFR_MDL),  // 0x9A
        -(int8_t)sizeof(MFR_REV),  // 0x9B
        -(int8_t)sizeof(MFR_LOC),  // 0x9C
        -(int8_t)sizeof(MFR_DAT),  // 0x9D
        -(int8_t)sizeof(MFR_SER),  // 0x9E
        1,                         // 0x3A
        2,                         // 0x3B
        2,                         // 0x3C
        1,                         // 0x3D
        2};                        // 0x3E

// only block write-block read process calls modify the length
static int8_t cmd_data_length_query = -1;

static cmd_t cmds_cmds[] = (cmd_t[]){
        {0x00, (int8_t *)&cmd_data_lengths[0], (uint8_t *)&page,     (fp_t)NULL, &page_chk,  (fp_t)NULL, QUERY_WR | QUERY_RD | QUERY_FMT_8B, 0},
        {0x1A, (int8_t *)&cmd_data_length_query, (uint8_t *)&query_r,  (fp_t)NULL, &query_prp, (fp_t)NULL, QUERY_SUP, 1},
        {0x8B, (int8_t *)&cmd_data_lengths[2], (uint8_t *)&curpage_volt, &accvolt, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x8C, (int8_t *)&cmd_data_lengths[3], (uint8_t *)&curpage_curr, &acccurr, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x8D, (int8_t *)&cmd_data_lengths[4], (uint8_t *)&temps_lin[0], (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x8E, (int8_t *)&cmd_data_lengths[5], (uint8_t *)&temps_lin[1], (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x8F, (int8_t *)&cmd_data_lengths[6], (uint8_t *)&temps_lin[2], (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x90, (int8_t *)&cmd_data_lengths[7], (uint8_t *)&frpms_lin[0], (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x91, (int8_t *)&cmd_data_lengths[8], (uint8_t *)&frpms_lin[1], (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x92, (int8_t *)&cmd_data_lengths[9], (uint8_t *)&frpms_lin[2], (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x96, (int8_t *)&cmd_data_lengths[10], (uint8_t *)&curpage_powr, &accpowr, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_LIN, 0},
        {0x99, (int8_t *)&cmd_data_lengths[11], (uint8_t *)&MFR_ID,   (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_NAN, 0},
        {0x9A, (int8_t *)&cmd_data_lengths[12], (uint8_t *)&MFR_MDL,  (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_NAN, 0},
        {0x9B, (int8_t *)&cmd_data_lengths[13], (uint8_t *)&MFR_REV,  (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_NAN, 0},
        {0x9C, (int8_t *)&cmd_data_lengths[14], (uint8_t *)&MFR_LOC,  (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_NAN, 0},
        {0x9D, (int8_t *)&cmd_data_lengths[15], (uint8_t *)&MFR_DAT,  (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_NAN, 0},
        {0x9E, (int8_t *)&cmd_data_lengths[16], (uint8_t *)&MFR_SER,  (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, QUERY_RD | QUERY_FMT_NAN, 0},
        {0x3A, (int8_t *)&cmd_data_lengths[17], (uint8_t *)&fan_config_1_2, (fp_t)NULL, &fan_config, (fp_t)NULL, QUERY_WR | QUERY_FMT_NAN, 0},
        {0x3B, (int8_t *)&cmd_data_lengths[18], (uint8_t *)&setfrpms_lin[0], (fp_t)NULL, &set_frpms, (fp_t)NULL, QUERY_WR | QUERY_FMT_LIN, 0},
        {0x3C, (int8_t *)&cmd_data_lengths[19], (uint8_t *)&setfrpms_lin[1], (fp_t)NULL, &set_frpms, (fp_t)NULL, QUERY_WR | QUERY_FMT_LIN, 0},
        {0x3D, (int8_t *)&cmd_data_lengths[20], (uint8_t *)&fan_config_3_4, (fp_t)NULL, &fan_config, (fp_t)NULL, QUERY_WR | QUERY_FMT_NAN, 0},
        {0x3E, (int8_t *)&cmd_data_lengths[21], (uint8_t *)&setfrpms_lin[2], (fp_t)NULL, &set_frpms, (fp_t)NULL, QUERY_WR | QUERY_FMT_LIN, 0}};

cmd_space_t cmds = {
        sizeof(cmds_cmds)/sizeof(cmd_t),
        cmds_cmds
};

void uc_reset();
void boot_new_fw();

static const int8_t ext_cmd_data_lengths[] = {
        1,  // 0x05
        1,  // 0x06
        4,  // 0xA0
        1,  // 0xB0
        13, // 0xC0
        7,  // 0xC1
        1}; // 0xC4

static cmd_t ext_cmds_cmds[] = (cmd_t[]){
        {0x05, (int8_t *)&ext_cmd_data_lengths[0], (uint8_t *)&dummy_byte, (fp_t)NULL, &boot_new_fw, (fp_t)NULL, 0, 0},
        {0x06, (int8_t *)&ext_cmd_data_lengths[1], (uint8_t *)&dummy_byte, (fp_t)NULL, &uc_reset, (fp_t)NULL, 0, 0},
        {0xA0, (int8_t *)&ext_cmd_data_lengths[2], (uint8_t *)&TMR_ERROR_CNT, (fp_t)NULL, (fp_t)NULL, (fp_t)NULL, 0, 0},
        {0xB0, (int8_t *)&ext_cmd_data_lengths[3], (uint8_t *)&use_pec_local, (fp_t)NULL, &set_pec, (fp_t)NULL, 0, 1},
        {0xC0, (int8_t *)&ext_cmd_data_lengths[4], (uint8_t *)&temp_curve_points_data, (fp_t)NULL, &set_tc_curve, (fp_t)NULL, 0, 0},
        {0xC1, (int8_t *)&ext_cmd_data_lengths[5], (uint8_t *)&temp_matrix_row, (fp_t)NULL, &set_tc_matrix, (fp_t)NULL, 0, 0},
        {0xC4, (int8_t *)&ext_cmd_data_lengths[6], (uint8_t *)&tc_on, (fp_t)NULL, &set_tc_onoff, (fp_t)NULL, 0, 0}};

cmd_space_t ext_cmds = {
        sizeof(ext_cmds_cmds)/sizeof(cmd_t),
        ext_cmds_cmds
};

static uint32_t fw_write_buf[FLASH_ROW_SIZE/4];

void uc_reset()
{
#ifdef USB_ENABLE
        usbdc_detach();
        usbdc_stop();
        usbdc_deinit();
#endif
        NVIC_SystemReset();
}


void boot_new_fw()
{
        uint32_t *char_user_flash = (uint32_t *)&user_flash;
        user_flash_t *tmp_user_flash = (user_flash_t *)&fw_write_buf;

        /* backup existing user data */
        for (uint8_t i = 0; i < FLASH_ROW_SIZE/4; ++i)
                fw_write_buf[i] = char_user_flash[i];

        /* set 0xBEC0ABCD copy FW code */
        tmp_user_flash->copy_fw = 0xbec0abcd;

        /* copy modified block back to the flash */
        flash_write_row((uint32_t *)&user_flash, (uint32_t *)fw_write_buf);

        /* reset the uC */
        uc_reset();
}


void __xMR page_chk()
{
        if (page > 2)
                page = 2;
}


void __xMR accvolt()
{
        curpage_volt = volts_lin[page];
}


void __xMR acccurr()
{
        curpage_curr = currs_lin[page];
}


void __xMR accpowr()
{
        curpage_powr = powrs_lin[page];
}


void query_prp()
{
        int16_t cmd_idx = in_addr_space(&cmds, query_r);
        if (cmd_idx != -1)
                query_r = QUERY_SUP | cmds.cmds[cmd_idx].query_byte;
        else
                query_r = 0;
}


void __xMR fan_config()
{
        fan_installed[0] = (fan_config_1_2 >> 7) & 0x1;
        fan_cmdrpm[0] = (fan_config_1_2 >> 6) & 0x1;
        fan_ppr[0] = (fan_config_1_2 >> 4) & 0x3;

        fan_installed[1] = (fan_config_1_2 >> 3) & 0x1;
        fan_cmdrpm[1] = (fan_config_1_2 >> 2) & 0x1;
        fan_ppr[1] = fan_config_1_2 & 0x3;

        fan_installed[2] = (fan_config_3_4 >> 7) & 0x1;
        fan_cmdrpm[2] = (fan_config_3_4 >> 6) & 0x1;
        fan_ppr[2] = (fan_config_3_4 >> 4) & 0x3;
}


void __xMR set_frpms()
{
        setfrpms[0] = linear_to_float(setfrpms_lin[0]);
        setfrpms[1] = linear_to_float(setfrpms_lin[1]);
        setfrpms[2] = linear_to_float(setfrpms_lin[2]);
}


void __xMR set_pec()
{
        use_pec = use_pec_local;
}


void __xMR set_tc_curve()
{
        uint8_t fan_n = temp_curve_points_data[0];
        for (uint8_t i = 0; i < 3; ++i) {
                temp_curve_points_x[fan_n][i] = linear_to_float(temp_curve_points_data[1 + 2*i]);
                temp_curve_points_y[fan_n][i] = linear_to_float(temp_curve_points_data[1 + 2*i + 1]);
        }
}


void __xMR set_tc_matrix()
{
        uint8_t fan_n = temp_matrix_row[0];
        for (uint8_t i = 0; i < 3; ++i) {
                temp_matrix[fan_n][i] = linear_to_float(temp_matrix_row[1 + i]);
        }
}


void __xMR set_tc_onoff()
{
        temp_control_on = tc_on;
}
